from .metadata import CRAIGSLIST


class Base:
    """ Base class for `Continent`, `Country`, `Site`, and `Area`. """

    _selector_key = ""
    _subclass_selector_key = ""
    _subclass = None

    def __init__(self):
        self._key = ""

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self._key)

    def __len__(self):
        return len(self._unique_class_keys(self._subclass_selector_key, self._key))

    def __iter__(self):
        """ Yield instance(s) of caller's subclass if it exists.
        Sublass instances are within scope of the caller's instance; for example,
        `Continent('africa')` will yield all `Country` instances that are in Africa. """
        return self._iter_class(self._subclass, self._subclass_selector_key, key=self._key)

    @classmethod
    def _iter_class(cls, _class, selector_key, key=""):
        """ Yield instances of `_class` instantiated by `key`, or if an empty string,
        all instances of `_class` instantiated by `selector_key`. """
        yield from (_class(class_key) for class_key in cls._unique_class_keys(selector_key, key))

    @classmethod
    def _unique_class_keys(cls, selector_key, key):
        """ Return a unique list of class instantiation `key`s as specified by `selector_key`.
        For example, if our `selector_key` is 'country' and  `key` is 'japan', we'll return ['japan'],
        rather than a list of eight `japan`s. If `key` is an empty string, return a unique list of keys
        for all values of `selector_key`["key"]. """
        return list(
            {
                region[selector_key]["key"]
                for region in cls._filter_tree(key)
                # determine if caller-specified `selector_key` in current instance's `_selector_key`
                if region.get(selector_key)
            }
        )

    @classmethod
    def _filter_tree(cls, key):
        """ Yield a sequence of dictionaries that fulfills the instance's filter as determined
        by `key`. For example, if we are in the `Continent` class, and `key` is 'africa', we'll
        yield a dict iterable where the 'continent' key has value 'africa' in `CRAIGSLIST`. """
        yield from filter(
            lambda leaf: leaf.get(cls._selector_key) and leaf[cls._selector_key]["key"] == key
            if key
            else leaf[cls._selector_key]["key"],
            CRAIGSLIST,
        )

    @classmethod
    def all(cls):
        """ Return all instances of the current class. """
        yield from cls._iter_class(cls, cls._selector_key)

    @property
    def key(self):
        """ Return key of the instance. """
        return self._key

    @property
    def title(self):
        """ Return title of the instance. """
        return self._search_tree("title")

    def _search_tree(self, key):
        """ Return value for `key` within the instance's filtered values. """
        try:
            return next(self._filter_tree(self._key))[self._selector_key].get(key)
        except StopIteration:
            return None



from .base import Base


class Area(Base):
    """ Parse Craigslist areas. """

    _selector_key = "area"

    def __init__(self, key):
        self._key = key

    @staticmethod
    def __iter__():
        # `Area` does not have a subclass
        yield from ()

    @property
    def url(self):
        """ Return craigslist url of `Area` instance. """
        return self._search_tree("url")

    @staticmethod
    def all():
        """ Unlike `Continent`, `Country`, and `Site`, `Area` does not
        have a subclass and therefore this method should be removed. """
        raise NotImplementedError("type object 'Area' has no attribute 'all'")


class Site(Base):
    """ Parse Craiglist sites. """

    _selector_key = "site"
    _subclass_selector_key = "area"
    _subclass = Area

    def __init__(self, key):
        self._key = key

    @property
    def url(self):
        """ Return craigslist url of Site instance. """
        return self._search_tree("url")

    def has_area(self):
        """ Boolean value for if site has areas. For example, `Site('sfbay')` has areas,
        `Site('monterey')` does not. """
        try:
            next(self.__iter__())
            return True
        except StopIteration:
            return False


class Country(Base):
    """ Parse Craiglist countries. """

    _selector_key = "country"
    _subclass_selector_key = "site"
    _subclass = Site

    def __init__(self, key):
        self._key = key


class Continent(Base):
    """ Parse Craiglist continents. """

    _selector_key = "continent"
    _subclass_selector_key = "country"
    _subclass = Country

    def __init__(self, key):
        self._key = key



import craigslist_meta
from setuptools import setup, find_packages


_classifiers = [
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Topic :: Software Development :: Libraries",
    "Topic :: Utilities",
]

if __name__ == "__main__":
    setup(
        name="python-craigslist-meta",
        version=craigslist_meta.__version__,
        author="Ira Horecka",
        author_email="ira89@icloud.com",
        url="https://github.com/irahorecka/python-craigslist-meta",
        py_modules=["craigslist_meta"],
        description="A simple API to traverse Craigslist endpoints",
        long_description=open("README.rst", encoding="utf-8").read(),
        license="MIT",
        classifiers=_classifiers,
        packages=find_packages(),
    )
